/*
 * This tool has been originally created by Christian Peters <Ciber DK>
 * 
 * modified by Karl Koch BNetzA 220-1 fon: +49 (0) 681 9330 515 email:
 * karl.koch@bnetza.de; karl.koch@ties.itu.int
 */

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.List;
import java.util.Vector;

import javax.swing.JFrame;
import javax.swing.JOptionPane;

import org.seamcat.model.plugin.*;
import org.seamcat.model.plugin.PostProcessingPlugin.ParameterType;

/**
 * 
 * @author Karl Koch, karl.koch@bnetza.de
 * @version <strong>For presentation to the SEAMCAT workshop June 2013</strong>
 *
 */
public class ShowCollectedData implements PostProcessingPlugin,
		ConsistencyCheck {

	private PluginDistribution defaultDistribution;
	
	private String filename = ""; // left empty as it is user dependent
	private String defaultPath = "";
	private String pathToSave; // combines defaultPath and filename
	private boolean shouldBeSaved = false; // flag whether the output file should be generated
	private Vector<String> outputFormat = null; // list for combobox output
	private int outputFormatIndex = 1; // sets .xml as default
	private Vector<Double> rLP; // collects the calculated location probabilities	
	private CollectedDataUnwantedAndBlocking singleResult = new CollectedDataUnwantedAndBlocking(); // new object containing the simulated values of a single shot
	private Vector<CollectedDataUnwantedAndBlocking> simulatedResults = null; // collects all singleResult
	private Transceiver rWt; // needed to record the distance between Vr and Wt
	private double[] oneShot; // used to copy not changed positions
	private boolean multipleLinks = false;	
	private int totalEvents;
	private String header;
	private String propagationModel;
	public BridgeToPostLoadDataR3 toReport; 
	private BufferedWriter out = null;
	ProgressBarStoreFileSE7 progressstore  = null;
	
/*	public double rCoverIplusN = 20; // in dB
	public double rCoverI = 20;
	public double rIoverN = 0;*/
	public double simulationRadius = 6.7; // in km
	public double locationProbability = 95.0; // in %
	public double maxDistance ;
	private double minDRSS = -86.15;// dRSS required for LP = 95%
	private double STDDEV = 5.5; // standard deviation
	private double blockingResponse;
	public int countVector;

/*	public CollectDataLoad_rev1() {
	}*/

	public void init(ParameterFactory param) {
		defaultDistribution = param.createDistribution(
				PluginDistribution.DistributionType.DiscreteUniform, 0, 360);
		/** fills the fields of the combobox */
		outputFormat = new Vector<String>();
		outputFormat.add(".txt");
		outputFormat.add(".xml");
		outputFormat.add("");

	}

	public void cleanUp() {
		// TODO Auto-generated method stub
	}

	public String getDescription() {
		return "This plug-in is to collect simulated data like victim's position and wanted signal";
	}

	public void process(ScenarioInfo scenario) throws Exception {
		int currentEvent = scenario.getCurrentEventNumber();
		if (scenario.getCurrentEventNumber() <= scenario.getNumberOfInterferingLinks()) {
			initPlugin(scenario);
		} else if (scenario.getCurrentEventNumber() == scenario.getTotalNumberOfEvents()) {
			if (toReport != null)
				prepareReport(scenario);
			else {
				toReport = new BridgeToPostLoadDataR3();
				prepareReport(scenario);
			}

			if (this.shouldBeSaved) {
				storeResults(scenario);
			}
		} else if (scenario.getCurrentEventNumber() > scenario.getNumberOfInterferingLinks()) {
			// TODO
			collectData(scenario);
		}

	}
	/**
	 * 
	 * @param scenario
	 * @throws Exception 
	 */
	private void storeResults(ScenarioInfo scenario) throws Exception {
		/** the extension of the file name defines the output format */
		String part = filename.substring(filename.indexOf(".") + 1);
		if (part.equals("txt"))
			saveResults(scenario); // as .txt file
		else if (part.equals("xml"))
			saveResultsXML(scenario);
		/*
		 * else if (part.equals("csv")) { saveResults(scenario); // as .csv file
		 * } // doesn't work correctly as the format is wrong
		 */
		else {
			// TODO if other extension should be added
		}
	}

	/**
	 * {@literal stores the generated table as text file }
	 * 
	 * @param scenario
	 */
	private void saveResults(ScenarioInfo scenario) {
		pathToSave = defaultPath + "\\" + filename;
		int fSize = (int) (scenario.getNumberOfInterferingLinks() * 6 * 20 * totalEvents / 1024.0 / 1024.0 + 1); // 
		ProgressBarStoreFile progressstore = new ProgressBarStoreFile("Saving TXT file in progress (about " + fSize
				+ " MB)");
		progressstore.init();
		progressstore.progressBar.setMinimum(0);
		progressstore.progressBar.setMaximum(countVector);
		try {
			BufferedWriter out = new BufferedWriter(new FileWriter(pathToSave));

			out.write(header); // header of the columns of the table 
			out.newLine();

			for (int i = 0; i < countVector; i++) {
				out.write(simulatedResults.get(i).event + ";");
				out.write(simulatedResults.get(i).x + ";");
				out.write(simulatedResults.get(i).y + ";");
				out.write(simulatedResults.get(i).dRSS + ";");
				out.write(simulatedResults.get(i).iRSS + ";");
				out.write(simulatedResults.get(i).noiseFloor + ";");
				out.write(simulatedResults.get(i).isMultipleILK + ";");
				out.write(simulatedResults.get(i).ilk + ";");
				out.write(simulatedResults.get(i).cOverN + "");
				out.newLine();
				progressstore.progressBar.setValue(i + 1);
			}
			out.flush();
			out.close();
			progressstore.stop();
		} catch (IOException e) {
			e.printStackTrace();
		}

	}

	/**
	 * {@literal saveResultsXML stores the result as XML file}
	 * @throws Exception 
	 */
	private void saveResultsXML(ScenarioInfo scenario) throws Exception {
		pathToSave = defaultPath + "\\" + filename;
		int fSize = (int) (scenario.getNumberOfInterferingLinks() * 12 * 32 * scenario.getTotalNumberOfEvents() / 1024.0 / 1024.0 + 1); // seven columns each 32 bytes
		progressstore = new ProgressBarStoreFileSE7("Saving XML file in progress (about " + fSize
				+ " MB)");
		progressstore.init();
		progressstore.progressBar.setMinimum(0);
		progressstore.progressBar.setMaximum(countVector);
		try {
			out = new BufferedWriter(new FileWriter(pathToSave));
			String xmlDeclaration = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?>";
			String xmlRoot = "result";
			String xmlElement = "event";

			out.write(xmlDeclaration);
			out.newLine();
			out.write("<" + xmlRoot + ">");
			out.newLine();
			for (int i = 0; i < countVector; i++) {
				String element = "<" + xmlElement;
				element += " eventNr = \"" + simulatedResults.get(i).event;
				element += " \" x = \"" + simulatedResults.get(i).x;
				element += " \" y = \"" + simulatedResults.get(i).y;
				element += " \" dRSS = \"" + simulatedResults.get(i).dRSS;
				element += " \" iRSSunw = \"" + simulatedResults.get(i).iRSS;
				element += " \" iRSSblock = \"" + simulatedResults.get(i).iRSSblock;
				element += " \" noiseFloor = \"" + simulatedResults.get(i).noiseFloor;
				element += " \" multiple = \"" + simulatedResults.get(i).isMultipleILK;
				element += " \" ilk = \"" + simulatedResults.get(i).ilk;
				element += " \" CoverI = \"" + scenario.getVictimLink().getReceiver().getRequiredCIRatio() ;
				element += " \" CoverNplusI = \"" + scenario.getVictimLink().getReceiver().getRequiredCNIRatio() ;
				element += " \" IoverN = \"" + scenario.getVictimLink().getReceiver().getRequiredINRatio();
				element += " \" minDRSS = \"" + minDRSS ;
				element += " \" blockingResponse = \"" + blockingResponse;
				element += " \" simulationRadius = \"" + simulationRadius;
				element += " \" model = \"" + propagationModel ;
				element += " \" stdDEV = \"" + STDDEV + "\">";
				element += "</" + xmlElement + ">";
				out.write(element);
				out.newLine();
				progressstore.progressBar.setValue(i + 1);
			}
			out.write("</" + xmlRoot + ">");
			out.flush();
			out.close();
			storeDTD(); // saves the corresponding xml namespace as result.dtd
			progressstore.stop();
		} catch (IOException e) {
			progressstore.frame.setVisible(false);
			showWrongPathMessage(defaultPath);
			defaultPath = "C:"; // valid path before re-starting this method
			saveResultsXML(scenario);
			// e.printStackTrace();
		}
	}

	private double getBlockingResponse(ScenarioInfo scenario, InterferingLink link) throws Exception {
		// TODO Auto-generated method stub
		blockingResponse = 0;
		Function2D blockingFunction = scenario.getVictimLink().getReceiver().getBlockingFunction();
		if (blockingFunction.toString().indexOf("Constant") != -1) {
			blockingResponse = blockingFunction.evaluate(0);
		} else {
			blockingResponse = blockingFunction.evaluate((link.getFrequency() - scenario.getVictimLink()
					.getFrequency()));
		}
		return blockingResponse;
	}

	private void showWrongPathMessage(String defaultPath) {
		// TODO Auto-generated method stub
		String wrongPath = defaultPath;

		defaultPath = "C:";
		pathToSave = defaultPath + "\\" + filename;
		String msg = "Due to a not existing path definition (" + wrongPath + ") \nthe location of the saved file has been changed to " + pathToSave;
		JFrame frame = new JFrame("Simulation results");
		frame.setResizable(true);
		frame.setSize(400, 400);
		JOptionPane.showMessageDialog(frame, msg, "File save warning", JOptionPane.WARNING_MESSAGE);
		progressstore.frame.setVisible(false);
	}

	/**
	 * {@literal storeDTD generates the DTD needed for the above XML file to that path the result has been saved}
	 */
	private void storeDTD() {
		try {
			BufferedWriter out = new BufferedWriter(new FileWriter(defaultPath + "\\resultSchema.dtd"));
			String DTD = "<?xml version=\"1.0\" encoding=\"UTF-8\"?> <!ELEMENT result ((event+))> <!ELEMENT event EMPTY>"
					+ "<!ATTLIST event	CoverI CDATA #REQUIRED	CoverNplusI CDATA #REQUIRED	IoverN CDATA #REQUIRED	dRSS CDATA #REQUIRED   "
					+ "	eventNr CDATA #REQUIRED	iRSSblock CDATA #REQUIRED	iRSSunw CDATA #REQUIRED	ilk CDATA #REQUIRED	minDRSS CDATA #REQUIRED	model CDATA #REQUIRED   "
					+ "	multiple CDATA #REQUIRED	noiseFloor CDATA #REQUIRED	simulationRadius CDATA #REQUIRED	blockingResponse CDATA #REQUIRED	 stdDEV CDATA #REQUIRED	x CDATA #REQUIRED	y CDATA #REQUIRED>";	

			out.write(DTD);
			out.flush();
			out.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}


	private void collectData(ScenarioInfo scenario) throws Exception {
		// TODO Auto-generated method stub
		readPosition(scenario);
		readDRSS(scenario);
		readIRSS(scenario);

	}

	/**
	 * {@literal adds the current event number and the positions x and y of the victim to the current row; }
	 * {@literal oneShot[] collects the values of event, x, y and dRSS which are not changed in case different interfering links have to be considered}
	 * 
	 * @param scenario
	 */
	private void readPosition(ScenarioInfo scenario) {
		oneShot = new double[4]; // reset to zero
		oneShot[0] = scenario.getCurrentEventNumber();
		oneShot[1] = scenario.getVictimLink().getReceiver().getPositionX();
		oneShot[2] = scenario.getVictimLink().getReceiver().getPositionY();
		double radius = scenario.getVictimLink().getReceiver()
				.getDistanceTo(rWt);
		if (radius > maxDistance)
			maxDistance = radius; // updates the max distance simulated; needed for the correct scaling of the plot 
	}

	/**
	 * {@literal adds the dRSS to the current row}
	 * 
	 * @param scenario
	 */
	private void readDRSS(ScenarioInfo scenario) {
		oneShot[3] = scenario.getVictimLink().getDRSSValue();
	}

	/**
	 * {@literal adds the iRSS and the index of the interfering link to the table; }
	 * {@literal in case more than one interfering link exist the values for event, x, y and dRSS are copied}
	 * 
	 * @param scenario
	 * @throws Exception
	 */
	private void readIRSS(ScenarioInfo scenario) throws Exception {
		List<InterferingLink> ilks = scenario.getInterferingLinks();
		if (scenario.getNumberOfInterferingLinks() == 1)
			multipleLinks = false;
		else
			multipleLinks = true;
		for (InterferingLink link : ilks) {
			singleResult = new CollectedDataUnwantedAndBlocking();
			singleResult.event = (int) oneShot[0];
			singleResult.x = oneShot[1];
			singleResult.y = oneShot[2];
			singleResult.dRSS = oneShot[3];
			singleResult.iRSS = link.getIRSSUnwanted();
			singleResult.iRSSblock = link.getIRSSBlocking();
			singleResult.blockingResponse = getBlockingResponse(scenario,link) ;
			singleResult.cOverI = 0;// not used as calculated by BridgeToPost 	
			singleResult.noiseFloor = getNoiseFloor(scenario);
			singleResult.isMultipleILK = multipleLinks;
			singleResult.ilk = link.getLinkIndex();
			singleResult.cOverN = 0; // not used as calculated by BridgeToPost 		
			singleResult.cIverN = 0; // not used as calculated by BridgeToPost 	
			simulatedResults.add(singleResult);
			countVector++; // Counter for the rows of the resulting table
		}

	}

	private double getNoiseFloor(ScenarioInfo scenario) {
		return scenario.getVictimLink().getReceiver().getNoiseFloor().trial();
	}

	private void prepareReport(ScenarioInfo scenario) {
		// informs the sub-class about the parameters changed by the simulation
		if(toReport == null) toReport = new BridgeToPostLoadDataR3();
		toReport.getFileChooser().setCurrentDirectory(
				(new File(scenario.getCopyOfWorkspace().getAbsoluteLocation())));
		toReport.setSimulatedResults(simulatedResults); 
		toReport.setSimulationRadius(simulationRadius);
		toReport.setMaxDistance(maxDistance);
		toReport.setMinDRSS(minDRSS);
		toReport.setSTDDEV(STDDEV);  
		toReport.setLocationProbability(locationProbability);  
		toReport.setScenario(scenario);
		toReport.setBlockingResponse(blockingResponse);
		toReport.initPlugin();
		toReport.setVisible();
		scenario.requestSimulationStop(); 
	}

	private void initPlugin(ScenarioInfo scenario) {
		// TODO Auto-generated method stub
		initWorkspace(scenario);
		
		simulatedResults = new Vector<CollectedDataUnwantedAndBlocking>(); // CollectedData of the data simulated
		
		countVector = 0; // vector to the rows of the table

		header = "event;x;y;dRSS;iRSS;noiseFloor;multiple;ilk;CoverN"; // used only for the .txt-output
		maxDistance = 0; // reset in case the simulation radius has been changed
		
		this.propagationModel = scenario.getVictimLink().getPropagationModel().getAssumedModel().name();

		if (outputFormatIndex == 0 && filename.indexOf(".") == -1)
			filename += ".txt";
		else if (outputFormatIndex == 1 && filename.indexOf(".") == -1)
			filename += ".xml";

		if (filename.indexOf(".") == -1) {// sets xml as default in case no extension selected
			outputFormatIndex = 1;
			filename += ".xml";
		}

	}

	/**
	 * 
	 * @param scenario
	 */
	private void initWorkspace(ScenarioInfo scenario) {
		rWt = scenario.getVictimLink().getTransmitter(); // needed to get the max distance
		double radius = scenario.getVictimLink().getReceiver()
				.getDistanceTo(rWt);
		if (radius > maxDistance)
			maxDistance = radius; // used to scale the plot	
	}

	public int getNumberOfParameters() {
		return 8;
	}

	public ParameterType getParameterType(int id) {

		switch (id) {
			case 1: {
				return ParameterType.String; // filename for the output
			}
			case 2: {
				return ParameterType.String; // default path for the output files
			}
			case 3: {
				return ParameterType.Integer; // outputFormatIndex
			}
			case 4: {
				return ParameterType.Boolean; // flag to save output
			}
			case 5: {
				return ParameterType.Double; // simulation radius
			}
			case 6: {
				return ParameterType.Double; // location probability
			}
			case 7: {
				return ParameterType.Double; // dRSS
			}
			case 8: {
				return ParameterType.Double; // STDDEV
			}
			default: {
				throw new IllegalArgumentException("illegal parameter id");
			}

		}
	}

	public void setParameterValue(int id, Object value) {
		try {
			switch (id) {
				case 1: {
					this.filename = (String) value;
					break;
				}
				case 2: {
					this.defaultPath = (String) value;
					break;
				}
				case 3: {
					this.outputFormatIndex = (Integer) value;
					break;
				}
				case 4: {
					this.shouldBeSaved = (Boolean) value;
					break;
				}
				case 5: {
					this.simulationRadius = (Double) value;
					break;
				}
				case 6: {
					this.locationProbability = (Double) value;
					break;
				}
				case 7: {
					this.minDRSS = (Double) value;
					break;
				}			
				case 8: {
					this.STDDEV = (Double) value;
					break;
				}
				default: {
					throw new IllegalArgumentException("illegal parameter id");
				}
			}
		} catch (Exception e) {

		}
	}

	public Object getParameterValue(int id) {
		switch (id) {
			case 1: {
				return this.filename;
			}
			case 2: {
				return this.defaultPath;
			}
			case 3: {
				return this.outputFormatIndex;
			}
			case 4: {
				return this.shouldBeSaved;
			}			
			case 5: {
				return this.simulationRadius;
			}
			case 6: {
				return this.locationProbability;
			}
			case 7: {
				return this.minDRSS;
			}
			case 8: {
				return this.STDDEV;
			}
			default: {
				throw new IllegalArgumentException("illegal parameter id");
			}

		}
	}

	public String getParameterName(int id) {
		switch (id) {
			case 1: {
				return "file name of the output data";
			}

			case 2: {
				return "default path ";
			}
			case 3: {
				return "output format ";
			}

			case 4: {
				return "save output file ";
			}
			case 5: {
				return "simulation radius  [km]";
			}

			case 6: {
				return "location probability [%]";
			}
			case 7: {
				return "minimum wanted Signal  [dBm]";
			}
			case 8: {
				return "standard deviation [dB]";
			}
			default: {
				throw new IllegalArgumentException("illegal parameter id");
			}

		}
	}

	/*
	 * check consistency of the workspace
	 */
	public boolean check(ScenarioInfo scenario) {
		boolean checkPropagationModel = true;
		String warning = "<html><br>";
		warning += "<br/><br/><strong>Post Processing Plugin: </strong>";
		warning += "<br />Please ensure that the values of <br/><strong>C/I, C/(I+N) and I/N</strong><br/>"
				+ "are defined as you want, even if there is still a consitency warning."
				+ "<br>";

		if (shouldBeSaved && !checkFilenameAndPath()) {
			filename = "outputData.xml";
			defaultPath = "C:";
			pathToSave = defaultPath + "\\" + filename;
			warning += "<br />Due to the location for the output files has not been defined,<br />the collected data will be saved onto <strong>"
					+ pathToSave
					+ "</strong> <br /> and the corresponding DTD onto <strong>C:\\resultSchema.dtd</strong>";

		}
		checkPropagationModel = false;
		warning += "</html>";

		if (checkPropagationModel == false) {
			scenario.addConsistencyWarning(warning);
			return false;
		} else {
			return true;
		}
	}

	@Override
	public Vector<String> getAllowedParameterValues(int index) {
		switch (index) {
			case 3:
				return this.outputFormat;
				/*
							case 5:
								return this.outputFigure;

							case 11:
								return this.criterion;

							case 13:
								return this.signalType;*/

		}
		return null;
	}

	/**
	 * {@literal tries to avoid an exception }
	 */
	private boolean checkFilenameAndPath() {
		if (filename.equals("") || defaultPath.equals("")) {
			return false;
		} else {
			return true;
		}
	}

}
